/*
 * add_cksum.c
 *
*/




#include "xkernel.h"
#include "add_cksum_i.h"
#include <krb5/krb5.h>
#include <isode/psap.h>
#include <krb5/kdb.h>
#include <krb5/asn1.h>
#include <com_err.h>
#include <krb5/asn.1/encode.h>

#ifdef __STDC__

void   add_cksum_init(XObj);
static xkern_return_t add_cksum_opendone(XObj , XObj, XObj, XObj);
static xkern_return_t add_cksum_demux(XObj, XObj, Msg *);
static xkern_return_t add_cksum_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t add_cksum_pop(XObj, XObj, Msg *, void *);
static xkern_return_t add_cksum_demux(XObj, XObj, Msg *);
static xkern_return_t add_cksum_control_pop(XObj, XObj, Msg *, void *);
static xmsg_handle_t add_cksum_push(XObj, Msg *);
static long mycopypop(void *, char*, long, void*);
static void mycopypush(void*, char*, long, void*);
static void add_cksum_control_sessn_init(XObj);
static void add_cksum_sessn_init(XObj);
static void mycopypush(void *, char *, long, void *);
static long mycopypop(void *, char *, long, void *);
static XObj CreateControlSessn(XObj, XObj, XObj, XObj);
static int add_cksum_control(XObj, int, char *, int);




#else

void   add_cksum_init();
static xkern_return_t add_cksum_opendone();
static xkern_return_t add_cksum_demux();
static xkern_return_t add_cksum_openenable();
static xkern_return_t add_cksum_pop();
static xkern_return_t add_cksum_demux();
static xmsg_handle_t add_cksum_push();
static xkern_return_t add_cksum_control_pop();
static long mycopypop();
static void mycopypush();
static void add_cksum_control_sessn_init();
static void add_cksum_sessn_init();
static void mycopypush();
static long mycopypop();
static XObj CreateControlSessn();
static int add_cksum_control();

#endif

int traceadd_cksump = 1;
#define ADDP add_cksump


void
add_cksum_init(self)
     XObj self;
{


  xTrace0(ADDP, TR_MAJOR_EVENTS, "add_cksum_init entered");
  self->state = xMalloc(sizeof(PSTATE));
  bzero((char*)self->state, sizeof(PSTATE));

  self->demux = add_cksum_demux;
  self->opendone = add_cksum_opendone;
  self->openenable = add_cksum_openenable;
  ((PSTATE*)self->state)->activemap = mapCreate(NUM_ENTRIES, sizeof(XObj));
  ((PSTATE*)self->state)->pasmap = mapCreate(NUM_ENTRIES, sizeof(XObj));

}



static xkern_return_t
add_cksum_openenable(self, hlpRcv, hlpType, p)
     XObj self, hlpRcv, hlpType;
     Part *p;
{
  PSTATE *ps = (PSTATE*)self->state;

  if ( defaultOpenEnable(ps->pasmap,hlpRcv, hlpType,
			 (VOID *)&hlpRcv) == XK_FAILURE ) {
    xTrace0(ADDP, TR_ERRORS, "add_cksum_openenable -- defaultOpenEnable failed");
    return XK_FAILURE;
  }

  if (ps->enable_count < 2) {
    xTrace1(ADDP, TR_EVENTS, "about to openenable down number %d",
	    ps->enable_count);
    return xOpenEnable(self, hlpRcv, xGetDown(self, ps->enable_count++), p);
  } else {
    xTrace0(ADDP, TR_ERRORS, "too many openenables received");
    return XK_FAILURE;
  }
}

static xkern_return_t
add_cksum_opendone(self, lls, llp, hlpType)
     XObj self, lls, llp, hlpType;
{
  PSTATE *ps = (PSTATE *)self->state;
  SSTATE *ss;
  XObj    sessn, control_sessn;
  Enable *e;

  if (self == hlpType) {
    xTrace0(ADDP, TR_ERRORS, "self == hlpType in add_cksum_opendone");
    return XK_FAILURE;
  }

  if (mapResolve(ps->pasmap, &hlpType, &e) == XK_FAILURE) {
    xTrace0(ADDP, TR_ERRORS,
	    "add_cksum_opendone: couldn't find enable object");
    return XK_FAILURE;
  }

  xDuplicate(lls);

  sessn = xCreateSessn(add_cksum_sessn_init, e->hlpRcv,
			 e->hlpType, self, 1, &lls);
  if (sessn == ERR_XOBJ) {
    xTrace0(ADDP, TR_ERRORS, "could not create session");
    return XK_FAILURE;
  }
  
  if (!ps->curr_session) {
    control_sessn = CreateControlSessn(e->hlpType, self, sessn, lls);
    if (control_sessn == ERR_XOBJ) {
      xTrace0(ADDP, TR_ERRORS, "could not create control session");
      return XK_FAILURE;
    }
    ps->curr_session = control_sessn;
  } else {
    control_sessn = ps->curr_session;
    ss = (SSTATE*)ps->curr_session->state;
    ss->upSessn[1] = sessn;
    ss->downSessn[1] = lls;
    ps->curr_session = 0;
  }

  mapBind(ps->activemap, &lls, control_sessn);
  return xOpenDone(e->hlpRcv, sessn, self);
}

    
static xkern_return_t
add_cksum_demux(self, lls, msg)
     XObj self, lls;
     Msg *msg;
{
  XObj control_sessn;
  xkern_return_t val;
  PSTATE  *ps = (PSTATE *)self->state;

  xTrace0(ADDP, TR_EVENTS, "add_cksum_demux -- entered");

  if (mapResolve(ps->activemap, &lls, &control_sessn) == XK_FAILURE) {
    xTrace1(ADDP, TR_EVENTS, "mapResolve failed for lls 0x%x", lls);
    return XK_FAILURE;
  }

  val = xPop(control_sessn, lls, msg, 0);

  xTrace0(ADDP, TR_EVENTS, "add_cksum_demux -- exiting");

  return val;
}


static XObj CreateControlSessn(hlpType, self, sessn, lls)
     XObj hlpType, self, sessn, lls;
{
  XObj new_s;
  SSTATE *ss;

  new_s = xCreateSessn(add_cksum_control_sessn_init,
		       hlpType, hlpType, self, 0, 0);
  if (new_s == ERR_XOBJ) {
    return new_s;
  }

  ss = X_NEW(SSTATE);
  bzero((char*)ss, sizeof(SSTATE));

  ss->upSessn[0] = sessn;
  ss->downSessn[0] = lls;

  new_s->state = (void*)ss;
  return new_s;
}

static void add_cksum_control_sessn_init(self)
     XObj self;
{
  self->pop = add_cksum_control_pop;
}

static void add_cksum_sessn_init(self)
     XObj self;
{
  self->pop = add_cksum_pop;
  self->push = add_cksum_push;
  self->control = add_cksum_control;
}



static xmsg_handle_t
add_cksum_push(s, msg)
     XObj s;
     Msg *msg;
{
  return xPush(xGetDown(s, 0), msg);
}


static xkern_return_t
add_cksum_pop(s, lls, msg, hdr)
     XObj s, lls;
     Msg *msg;
     void *hdr;
{
    xTrace0(ADDP, TR_EVENTS, "add_cksum_pop - entered");
    return xDemux(s, msg);
}
  
static int add_cksum_control(self, opcode, buf,len)
     XObj self;
     int opcode, len;
     char *buf;
{
  xTrace0(ADDP, TR_EVENTS, "add_cksum_control entered");
  return xControl(xGetDown(self,0), opcode, buf, len);
}


static xkern_return_t
add_cksum_control_pop(self, lls, msg, hdr)
     XObj self, lls;
     Msg *msg;
     void *hdr;
{
  SSTATE              *ss = (SSTATE *)self->state;
  krb5_data            scratch;
  krb5_authenticator  *authent;
  int                  retval;
  xkern_return_t       val;

  xTrace0(ADDP, TR_MAJOR_EVENTS, "add_cksum_control_pop -- entered");

  if (lls == ss->downSessn[0]) {
    if (ss->msgFull[0]) {
      xTrace0(ADDP, TR_ERRORS, "two messages received from downSessn[0]");
      ss->msgFull[0] = 0;
      msgDestroy(&ss->msg[0]);
      return XK_FAILURE;
    }

    msgConstructCopy(&(ss->msg[0]), msg);
    ss->msgFull[0] = 1;
    return XK_SUCCESS;
  } else if (lls == ss->downSessn[1]) {
    /* there is much work to be done here */

    if ( !ss->msgFull[0] ) {
      xTrace0(ADDP, TR_ERRORS, "did not receive message from downSessn[0]");
      return XK_FAILURE;
    }

    /* outline  --   msgPop msg[0] into data; decode auth;
       push checksum onto msg[1]; Pop both messages up  
    */
    scratch.length = msgLen(&ss->msg[0]);
    scratch.data = (char*)xMalloc(scratch.length);
    if ( !scratch.data ) {
      xTrace0(ADDP, TR_ERRORS, "malloc failed");
      ss->msgFull[0] = 0;
      msgDestroy(&ss->msg[0]);
      return XK_FAILURE;
    }

    msgPop(&ss->msg[0], mycopypop, (char*)scratch.data, scratch.length,0);

    retval = decode_krb5_authenticator(&scratch, &authent);
    if (retval) {
      xTrace1(ADDP, TR_ERRORS, "error (%d) while decoding authenticator",
	      retval);
      ss->msgFull[0] = 0;
      msgDestroy(&ss->msg[0]);
      return XK_FAILURE;
    }

    msgPush(msg, mycopypush, (char *)authent->checksum->contents,
	    authent->checksum->length, 0);

    msgDestroy(&ss->msg[0]);
    ss->msgFull[0] = 0;
    msgConstructBuffer(&ss->msg[0], (char*)authent,
		       sizeof(krb5_authenticator));

    xPop(ss->upSessn[0], ss->downSessn[0], &ss->msg[0], 0);
    msgDestroy(&ss->msg[0]);

    val = xPop(ss->upSessn[1], ss->downSessn[1], msg, 0);

    xTrace0(ADDP, TR_EVENTS, "add_cksum_control_pop -- exiting");

    return val;
  } else {
    xTrace0(ADDP, TR_ERRORS, "message received from unrecognized lls");
    return XK_FAILURE;
  }
}

static long mycopypop(dest, src, count, arg)
     void *dest;
     char *src;
     long count;
     void *arg;
{
  bcopy(src, (char*)dest, count);
  return count;
}

static void mycopypush(src, dst, count, arg)
     void *src;
     char *dst;
     long  count;
     void *arg;
{
  bcopy((char *)src, dst, count);
}

