#include "xkernel.h"
#include "dheckxleaf.h"
#include "dheckxleaf_i.h"
#include <stdlib.h>
#include <memory.h>

int dheckxleafp;
int tracedheckxleafp;

void dheckxleaf_init( XObj );
static XObj		dheckxleafOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxleafOpenDone( XObj, XObj, XObj, XObj );
static xkern_return_t	dheckxleafPop( XObj, XObj, Msg *, void * );
static xkern_return_t	dheckxleafDemux( XObj, XObj, Msg * );
static void		dheckxleafSessnInit(XObj);
static xkern_return_t	dheckxleafClose( XObj );
static int		dheckxleafControl(XObj, int, char *, int);

void DheckxKnown(unsigned long *,unsigned long *, unsigned long *);
void Dheckx(unsigned long *,unsigned long *, unsigned long *,
		unsigned long *, unsigned long *);

static int
dheckxleafControl(self, opcode, buf, len)
    XObj self;
    int opcode;
    char *buf;
    int len;
{
   SState	*ss = (SState *)self->state;
   switch (opcode) {
	case DHECKXLEAF_PUBLIC_VALUE:
		{
			unsigned long * value = (unsigned long *) buf;
			int x;
    			if (len != 40)
				return XK_FAILURE;
			for(x = 0; x < DEGW; x++)
			{
				value[x] = htonl(ss->x[x]);
				value[x+DEGW] = htonl(ss->y[x]);
			}
			return XK_SUCCESS;
		}
		break;
	case DHECKXLEAF_CALCULATE_SECRET:
		{
  			unsigned long rx[DEGW];
 			unsigned long ry[DEGW];
  			unsigned long ix[DEGW];
 			unsigned long iy[DEGW];
			unsigned long * value = (unsigned long *) buf;
			int x;
   			if (len != 40)
				return XK_FAILURE;
			for(x = 0; x < DEGW; x++)
			{
				ix[x] = ntohl(value[x]);
				iy[x] = ntohl(value[x+DEGW]);
			}
			Dheckx(ix,iy,ss->expo,rx,ry);
			for(x = 0; x < DEGW; x++)
			{
				value[x] = htonl(rx[x]);
				value[x+DEGW] = htonl(ry[x]);
			}
			return XK_SUCCESS;
		}
		break;	
	case DHECKXLEAF_VALUE_SIZE:
		{
   			if (len != sizeof(int))
				return XK_FAILURE;
			*((int *)buf)=320;
			return XK_SUCCESS;
		}
		break;
      	default:
	return XK_FAILURE;
    }
}

static xkern_return_t dheckxleafDemux(self, lls, msg)
     XObj self, lls;
     Msg  *msg;
{
	return XK_SUCCESS;
}

static xkern_return_t dheckxleafPop(self, lls, msg, hdr) 	
	XObj  self,lls;
	Msg * msg;
	void * hdr;
{
	return XK_SUCCESS;
}

static xkern_return_t	dheckxleafOpenDone( self, lls, llp, hlpType  )
     XObj self, lls, llp, hlpType;
{
	return XK_SUCCESS;
}

static XObj dheckxleafOpen(self, hlp, hlpType, part)
     XObj self, hlp, hlpType;
     Part *part;
{
    	SState  * ss;
	XObj s;
	int x;
   	s = xCreateSessn(dheckxleafSessnInit, hlp, hlp, self, 0 , NULL);
	ss =(SState *) xMalloc(sizeof(SState));
    	bzero((void *)ss, sizeof(SState));
	for(x=0; x < DEGW-1; x++)
		ss->expo[x] = rand();
	ss->expo[DEGW-1] = 0;
	DheckxKnown(ss->expo,ss->x,ss->y);
	s->state = ss;
	return s;
}

static xkern_return_t
dheckxleafClose( self )
    XObj	self;
{
	memset(self->state,0,sizeof(SState));
	xFree(self->state);
	xDestroy(self);
	return XK_SUCCESS;
}



void dheckxleaf_init( self )
     XObj self;
{

  	self->open = dheckxleafOpen;
  	self->control = dheckxleafControl;
  	self->opendone = dheckxleafOpenDone;
  	self->pop = dheckxleafPop;
  	self->demux = dheckxleafDemux;

}

static void
dheckxleafSessnInit(s)
    XObj s;
{ 
 	s->close = dheckxleafClose;
  	s->pop = dheckxleafPop;
	s->control = dheckxleafControl;
}
