/*     
 * $RCSfile: consul_timing.c,v $
 *
 * x-kernel v3.2
 *
 * Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 *
 * $Log: consul_timing.c,v $
 * Revision 1.2  1994/05/06 20:22:44  menze
 *   [ 1994/05/05          menze ]
 *   Fixed prototype warning
 *   Use msgShow rather than MSG_SHOW
 *
 * Revision 1.1  1994/04/18  22:02:03  slm
 * Initial revision
 *
 */

/*
 * consul_timing.c
 * A modification of psync_timing.c to work on the Consul protocol graph
 * of Composer (Restart), Divider, Dispatch and Total_order.
 *
 * Configure the number and hosts of the participants by setting the array
 * part_index (e.g., int part_index[] = {FRANKLIN, KADMOS, FRANKLIN}; would
 * have part[0] on fra, part[1] on kad, and part[2] on fra).  See the
 * declaration of it and the corresponding comments below.  Also, see the
 * command line options parsed - the number of trips/test, the number of tests
 * and a debug value that the user code could use and use to set psync's
 * tracepsyncp (caution - this would set the trace value for all psync users on
 * a host, not just this test) can be set.
 */

#include "xkernel.h"
#include "psync.h"

/* from /etc/ethers */                                 /* shouldn't 2nd # be? */
#define CHE_ENET_ADDRESS 	{ 0x0800, 0x2001, 0xC3E4 }   /* 2008 */
#define JOJOBA_ENET_ADDRESS	{ 0x0800, 0x2001, 0xF716 }   /* 200D */
#define SAGUARO_ENET_ADDRESS	{ 0x0800, 0x2001, 0xC6E9 }   /* 2009 */
#define PALOVERDE_ENET_ADDRESS	{ 0x0800, 0x2001, 0xBE4A }   /* 2009 */
#define JAGUARUNDI_ENET_ADDRESS	{ 0x0800, 0x2001, 0x5CAD }   /* 2008 */
#define CHOLLA_ENET_ADDRESS	{ 0x0800, 0x2001, 0xB945 }   /* 200B */
#define OWL_ENET_ADDRESS     	{ 0x0800, 0x2001, 0xAC73 }   /* 2009 */
#define LANTANA_ENET_ADDRESS	{ 0x0800, 0x2001, 0xA735 }   /* 2008 */
#define ROADRUNNER_ENET_ADDRESS	{ 0x0800, 0x2001, 0x3B77 }   /* 200C */
#define EAGLE_ENET_ADDRESS	{ 0x0800, 0x2001, 0x2FBF }   /* 2008 */
#define HEDGEHOG_ENET_ADDRESS	{ 0x0800, 0x2001, 0x40C2 }   /* 200? */
#define JAVELINA_ENET_ADDRESS	{ 0x0800, 0x2001, 0xE826 }   /* 200F */
#define OPUNTIA_ENET_ADDRESS	{ 0x0800, 0x2001, 0x40C6 }   /* 200? */
#define BOBCAT_ENET_ADDRESS	{ 0x0800, 0x2001, 0x88BC }   /* 2009 */
#define ALOE_ENET_ADDRESS	{ 0x0800, 0x2001, 0x3B44 }   /* 200? */

/* from /etc/hosts */
#define CHE_IP_ADDRESS		{ 192, 12, 69, 60 }
#define JOJOBA_IP_ADDRESS	{ 192, 12, 69, 12 }
#define SAGUARO_IP_ADDRESS	{ 192, 12, 69, 13 }
#define PALOVERDE_IP_ADDRESS	{ 192, 12, 69, 16 }
#define JAGUARUNDI_IP_ADDRESS	{ 192, 12, 69, 30 }
#define CHOLLA_IP_ADDRESS	{ 192, 12, 69, 59 }
#define OWL_IP_ADDRESS		{ 192, 12, 69, 61 }
#define LANTANA_IP_ADDRESS	{ 192, 12, 69, 62 }
#define ROADRUNNER_IP_ADDRESS	{ 192, 12, 69, 65 }
#define EAGLE_IP_ADDRESS	{ 192, 12, 69, 66 }
#define HEDGEHOG_IP_ADDRESS	{ 192, 12, 69, 67 }
#define JAVELINA_IP_ADDRESS	{ 192, 12, 69, 68 }
#define OPUNTIA_IP_ADDRESS	{ 192, 12, 69, 69 }
#define BOBCAT_IP_ADDRESS	{ 192, 12, 69, 70 }
#define ALOE_IP_ADDRESS		{ 192, 12, 69, 71 }

#define CHE_PSYNC_ADDRESS		{ 2000, CHE_IP_ADDRESS }
#define JOJOBA_PSYNC_ADDRESS		{ 2001, JOJOBA_IP_ADDRESS }
#define SAGUARO_PSYNC_ADDRESS		{ 2002, SAGUARO_IP_ADDRESS }
#define PALOVERDE_PSYNC_ADDRESS		{ 2003, PALOVERDE_IP_ADDRESS }
#define JAGUARUNDI_PSYNC_ADDRESS	{ 2004, JAGUARUNDI_IP_ADDRESS }
#define CHOLLA_PSYNC_ADDRESS		{ 2005, CHOLLA_IP_ADDRESS }
#define OWL_PSYNC_ADDRESS    		{ 2006, OWL_IP_ADDRESS }
#define LANTANA_PSYNC_ADDRESS		{ 2007, LANTANA_IP_ADDRESS }
#define ROADRUNNER_PSYNC_ADDRESS	{ 2008, ROADRUNNER_IP_ADDRESS }
#define EAGLE_PSYNC_ADDRESS		{ 2009, EAGLE_IP_ADDRESS }
#define HEDGEHOG_PSYNC_ADDRESS		{ 2010, HEDGEHOG_IP_ADDRESS }
#define JAVELINA_PSYNC_ADDRESS		{ 2011, JAVELINA_IP_ADDRESS }
#define OPUNTIA_PSYNC_ADDRESS		{ 2012, OPUNTIA_IP_ADDRESS }
#define BOBCAT_PSYNC_ADDRESS		{ 2013, BOBCAT_IP_ADDRESS }
#define ALOE_PSYNC_ADDRESS		{ 2014, ALOE_IP_ADDRESS }

/* these correspond to indices in host_tab */
#define CHE        0
#define JOJOBA     1
#define SAGUARO    2
#define PALOVERDE  3
#define JAGUARUNDI 4
#define CHOLLA     5
#define OWL        6
#define LANTANA    7
#define ROADRUNNER 8
#define EAGLE      9
#define HEDGEHOG   10
#define JAVELINA   11
#define OPUNTIA    12
#define BOBCAT     13
#define ALOE       14

#define MESSAGE "xxxxxxxxyyyy3456789012345678901234567890123456"
#define MESSAGE_LENGTH 46	/* length of MESSAGE */

#define argPrefix(str) (!strncmp(globalArgv[i], str, strlen(str)))
#define argEq(str) (!strcmp(globalArgv[i], str))

       void           consul_timing_init(XObj);
static void           userProcess(Event, VOID *);
/* must be MAX_PARTS of the following declarations */
static xkern_return_t user_demux0(XObj, XObj, Msg *),
		      user_demux1(XObj, XObj, Msg *),
		      user_demux2(XObj, XObj, Msg *);
static xkern_return_t user_open_done(XObj, XObj, XObj, XObj);
static void           subtime(XTime *, XTime *, XTime *);
static long           stripMsg(VOID *, char *, long, VOID *);

static xkern_return_t (*user_demuxes[MAX_PART])( XObj, XObj, Msg * ) = {
    user_demux0, user_demux1, user_demux2
};

/* Global variable */
int traceconsul_timingp = 1;

/* These may be reset by command line options. */
static unsigned trips_per_test = 20;
static unsigned trip_num       = 20;
static unsigned num_tests      = 5;
static int      debug_val      = 0;

static unsigned test_num = 1;	/* current test */
static XObj     myProtl;
static XTime    starttime;

struct psync_host {
    char      *name;
    ETHhost   enet_address;
    PSYNCaddr psync_address;
};

static struct psync_host host_tab[] = {
    { "che", CHE_ENET_ADDRESS, CHE_PSYNC_ADDRESS },
    { "jojoba", JOJOBA_ENET_ADDRESS, JOJOBA_PSYNC_ADDRESS },
    { "saguaro", SAGUARO_ENET_ADDRESS, SAGUARO_PSYNC_ADDRESS },
    { "paloverde", PALOVERDE_ENET_ADDRESS, PALOVERDE_PSYNC_ADDRESS },
    { "jaguarundi", JAGUARUNDI_ENET_ADDRESS, JAGUARUNDI_PSYNC_ADDRESS },
    { "cholla", CHOLLA_ENET_ADDRESS, CHOLLA_PSYNC_ADDRESS },
    { "owl", OWL_ENET_ADDRESS, OWL_PSYNC_ADDRESS },
    { "lantana", LANTANA_ENET_ADDRESS, LANTANA_PSYNC_ADDRESS },
    { "roadrunner",  ROADRUNNER_ENET_ADDRESS,  ROADRUNNER_PSYNC_ADDRESS },
    { "eagle", EAGLE_ENET_ADDRESS, EAGLE_PSYNC_ADDRESS },
    { "hedgehog", HEDGEHOG_ENET_ADDRESS, HEDGEHOG_PSYNC_ADDRESS },
    { "javelina", JAVELINA_ENET_ADDRESS, JAVELINA_PSYNC_ADDRESS },
    { "opuntia", OPUNTIA_ENET_ADDRESS, OPUNTIA_PSYNC_ADDRESS },
    { "bobcat", BOBCAT_ENET_ADDRESS, BOBCAT_PSYNC_ADDRESS },
    { "aloe", ALOE_ENET_ADDRESS, ALOE_PSYNC_ADDRESS }
};

/* part_index is how you configure your test to run - just put the indeces
 * from host_tab into this, and you will get that many participants running on
 * those hosts.  Participant 0 is the one who does the active open, so
 * host_tab[part_index[0]].name should be booted last.  Note that the number in
 * this array must be <= MAX_PART, and there must be MAX_PART user_demuxN funcs
 * at the bottom.  (For my programs MAX_PART = NUM_PARTS - dhan)
 */
static int part_index[] = { CHE, JOJOBA, SAGUARO };

#define NUM_PARTS (sizeof(part_index) / sizeof(int))

/* This macro enables the hosts to respond to msgs in a round-robin fashion */
#define REPLY_TO_MSG(m,part_num) ((*((int *)m) % NUM_PARTS) == part_num ? 1 : 0)
/* #define REPLY_TO_MSG(m,part_num) 1 */

#define MSG_SHOW	msgShow

void consul_timing_init(self)
XObj self;
{
    int     i, rem;
    ETHhost myeaddr;
    IPhost  myipaddr;

    xTrace0(consul_timingp, TR_GROSS_EVENTS,
	    "USER: BEGIN: consul_timing timing test");
    myProtl = self;
    if (!xIsProtocol(xGetDown(myProtl, 0)))
	Kabort("CONSUL_TIMING: could not get lower protocol");

    /* Process Options */
    for (i = 1; i < globalArgc; i++) {
	if (argEq("-t"))
	    trip_num = trips_per_test = atoi(globalArgv[++i]);
	else if (argPrefix("-t"))
	    trip_num = trips_per_test = atoi(globalArgv[i] + 2);
	else if (argEq("-n"))
	    num_tests = atoi(globalArgv[++i]);
	else if (argPrefix("-n"))
	    num_tests = atoi(globalArgv[i] + 2);
	else if (argEq("-d"))
	    debug_val = atoi(globalArgv[++i]);
	else if (argPrefix("-d"))
	    debug_val = atoi(globalArgv[i] + 2);
	else {
            xError(
	    "usage: bxpsync [-t trips_per_test] [-n num_tests] [-d debug_val]");
	    return;
	}
    }

    /* print my ethernet address */
    if (xControl(xGetProtlByName("eth"), GETMYHOST, (char *)&myeaddr,
		 sizeof(ETHhost)) == -1)
	Kabort("USER: Cannot get my own enet address");
    xTrace3(consul_timingp, TR_ALWAYS,
	    "\nUSER: My ethernet address = %x.%x.%x\n",
	    myeaddr.high, myeaddr.mid, myeaddr.low);

    /* the number of tests must be divisible by NUM_PARTS */
    rem = trips_per_test % NUM_PARTS;
    if (rem) {
	trips_per_test += (NUM_PARTS - rem);   /* make divisible by NUM_PARTS */
	trip_num = trips_per_test;
	xTrace2(consul_timingp, TR_ALWAYS,
	    "USER: Really doing %d trips - must be divisible by NUM_PARTS (%d)",
		trips_per_test, NUM_PARTS);
    }
    if (NUM_PARTS > MAX_PART) {
	xTrace2(consul_timingp, TR_DETAILED,
		"USER: NUM_PARTS (%d) > MAX_PART (%d) - fix this",
	        NUM_PARTS, MAX_PART);
	Kabort("");
    }

    /*
     * Create a (fake) user protocol for each participant.  Participant 0 is
     * the one who will call the active open, so he should be called last
     * and his host should be booted last.
     */
    if (xControl(xGetProtlByName("ip"), GETMYHOST, (char *)&myipaddr,
	         sizeof(IPhost)) == -1)
	Kabort("USER: Cannot get my own IP address");
    for (i = NUM_PARTS - 1; i >= 0; i--) {
	myeaddr = host_tab[part_index[i]].enet_address;
	xTrace4(consul_timingp, TR_ALWAYS, "USER: host_tab index %d %d.%d.%d",
	        part_index[i], myeaddr.high, myeaddr.mid, myeaddr.low);

	/* Only fire up participants that belong on this host. */
	if (IP_EQUAL(myipaddr, host_tab[part_index[i]].psync_address.host)) {
	    /* create a protocol for participant i on this host */
	    xTrace1(consul_timingp, TR_ALWAYS,
		    "USER: firing up participant %d...", i);
	    evDetach(evSchedule(userProcess, (VOID *)i, 0));
	    xTrace1(consul_timingp, TR_ALWAYS,
		    "USER: consul_timing_init: created the protocols for %s",
		    host_tab[part_index[i]].name);
	    xTrace1(consul_timingp, TR_ALWAYS,
		    "USER: done firing up participant %d", i);
	}
    }
    xTrace0(consul_timingp, TR_ALWAYS,
	   "USER: DONE firing up all participants that belong on this host\n");
}

/*
 * Called when the event for the user processes is created.
 * Includes common open functions for all participants.
 */
static void userProcess(ev, arg)
Event ev;
VOID  *arg;
{
    int  part_num = (int)arg;
    Part part[MAX_PART+3];	/* + 3 so server can hold opcode, port,
			    	 *     all participants and NULL */
    int  opcode = 1;		/* everyone should have the same opcode
				 * for this program (see COMPOSER) */
    int  port = 1000;
    int  i;
    XObj mysessn;
    Msg  dg;
    char buf[MESSAGE_LENGTH];

    xTrace0(consul_timingp, TR_FUNCTIONAL_TRACE, "USER: entered userProcess");
    xTrace2(consul_timingp, TR_DETAILED, "USER: Participant %d = %s", part_num,
	   ipHostStr(&host_tab[part_index[part_num]].psync_address.host));
    myProtl->opendone = user_open_done;
    myProtl->demux    = user_demuxes[part_num];

    /* if part_num == 0
     *   partlist = [opcode,port,psync0,psync1,psync2,...,psyncN]
     *   do an active open
     *   prepare the first msg
     *   send the msg
     * else
     *   partlist = [opcode,port,my_psync_addr]
     *   do passive open
     */
    if (part_num == 0) {
        partInit(part, MAX_PART + 3);
        partPush(part[0], (VOID *)&opcode, sizeof(int));
        partPush(part[1], (VOID *)&port, sizeof(int));
	partPush(part[2], (VOID *)&host_tab[part_index[0]].psync_address,
		 sizeof(PSYNCaddr));
	for (i = 1; i < NUM_PARTS; i++) {
	    xTrace2(consul_timingp, TR_DETAILED, "USER: Participant %d = %s",
		   i, ipHostStr(&host_tab[part_index[i]].psync_address.host));
	    partPush(part[i+2],
		     (VOID *)&host_tab[part_index[i]].psync_address,
		     sizeof(PSYNCaddr));
	}
	partPush(part[NUM_PARTS+2], (VOID *)NULL, 0);
	xTrace0(consul_timingp, TR_DETAILED,
		"USER: Participant number 0 - open");
	if ((mysessn = xOpen(myProtl, myProtl, xGetDown(myProtl,0), part)) ==
	        ERR_XOBJ) {
	    xError("USER: Not sending, other host(s) down.");
	    xError("*** BOOT ME LAST!!!! ***");
	    return;
	}

	/*
	 * The first word of the message is the sequence number.  Currently,
	 * each participant replies to the message if the sequence number is
	 * equal to it modula NUM_PARTS.  This reply policy is fixed in
	 * REPLY_TO_MSG and thus can be easily modified.
	 */
	xTrace3(consul_timingp, TR_DETAILED,
		"USER: Sending %d trips per test (test %d of %d)...",
	        trips_per_test, test_num, num_tests);
	strcpy(buf, MESSAGE);
        *((int *)buf)   = 1;	/* first word is the count */
        *((int *)buf+1) = 0;	/* second word is sending participant */
        trip_num--;
	xGetTime(&starttime);
	msgConstructBuffer(&dg, buf, MESSAGE_LENGTH);
	xTrace2(consul_timingp, TR_DETAILED,
		"USER: Sending 1st msg, TIME %d.%d",
		starttime.sec, starttime.usec);
	xIfTrace(consul_timingp, TR_DETAILED) {
	    MSG_SHOW(&dg);
	}
	xPush(mysessn, &dg);
	msgDestroy(&dg);
    }
    else {	/* part_num > 0 */
        partInit(part, 3);
        partPush(part[0], (VOID *)&opcode, sizeof(int));
        partPush(part[1], (VOID *)&port, sizeof(int));
	partPush(part[2],
		 (VOID *)&host_tab[part_index[part_num]].psync_address,
		 sizeof(PSYNCaddr));
	xTrace1(consul_timingp, TR_DETAILED,
		"USER: Participant number %d - openenable\n", part_num);
	if (xOpenEnable(myProtl, myProtl, xGetDown(myProtl,0), part) ==
	        XK_FAILURE) {
	    xError("USER: couldn't openenable lower protocol");
	    return;
	}
    }
}

#ifdef PROFILE
#include "profile.c"
#endif

/*
 * Called when the 0th participant receives a msg from DISPATCH.
 */
static xkern_return_t user_demux0(self, lls, dg)
XObj self, lls;
Msg  *dg;
{
    XTime           now, total;
    static unsigned count = 0;
    char            buf[MESSAGE_LENGTH];
    Msg             rdg;

    xTrace1(consul_timingp, TR_FUNCTIONAL_TRACE,
	    "USER: entered user_demux0: trip_num = %d", trip_num);
    xGetTime(&now);		/* TEMP */
    xTrace3(consul_timingp, TR_DETAILED,
	    "USER: user_demux0: got msg of len %d at %d.%d",
            msgLen(dg), now.sec, now.usec);
    xIfTrace(consul_timingp, TR_DETAILED) {
	MSG_SHOW(dg);
    }
    if (trip_num > 1) {
	trip_num--;
	msgPop(dg, stripMsg, (VOID *)buf, MESSAGE_LENGTH, (VOID *)NULL);
	xTrace3(consul_timingp, TR_DETAILED,
	        "USER: seqno = %d, partno = %d, will reply = %d",
	        *((int *)buf), *((int *)buf+1), REPLY_TO_MSG(buf, 0));
	/*
	 * if my turn to reply
	 *   create a new msg
	 *   push it to DISPATCH
	 */
	if (REPLY_TO_MSG(buf, 0)) {
	    /* increment the sequence number and reply */
	    *((int *)buf)  += 1;	/* increment count */
            *((int *)buf+1) = 0;	/* insert my participant number */
            trip_num--;			/* don't receive my own message */
	    msgConstructBuffer(&rdg, buf, MESSAGE_LENGTH);
	    xTrace2(consul_timingp, TR_DETAILED,
		    "USER: Sending msg, TIME %d.%d", now.sec, now.usec);
	    xIfTrace(consul_timingp, TR_DETAILED) {
	        MSG_SHOW(&rdg);
	    }
	    xPush(lls, &rdg);
	    msgDestroy(&rdg);
	}
    }
    else {
	count += trips_per_test;
	subtime(&starttime, &now, &total);
	xTrace3(consul_timingp, TR_DETAILED,
		"USER: TIME %d.%d (%d total trips)",
		total.sec, total.usec, count);
	if (++test_num <= num_tests) {
	    /* start another test */
	    xTrace3(consul_timingp, TR_DETAILED,
		    "USER: Sending %d trips per test (test %d of %d)...",
	            trips_per_test, test_num, num_tests);
	    strcpy(buf, MESSAGE);
	    *((int *)buf)   = 1;	/* set count to zero */
            *((int *)buf+1) = 0;	/* my patricipant number */
	    trip_num = trips_per_test - 1;
	    xGetTime(&starttime);
	    msgConstructBuffer(&rdg, buf, MESSAGE_LENGTH);
	    xTrace2(consul_timingp, TR_DETAILED,
		    "USER: Sending msg, TIME %d.%d", now.sec, now.usec);
	    xIfTrace(consul_timingp, TR_DETAILED) {
	        MSG_SHOW(&rdg);
	    }
	    xPush(lls, &rdg);
	    msgDestroy(&rdg);
	}
        else
	    xTrace1(consul_timingp, TR_ALWAYS,
		    "USER: All done with the %d tests!!\n", num_tests);
    }
    return XK_SUCCESS;
}

/*
 * Called when the 1st participant receives a msg from DISPATCH.
 */
static xkern_return_t user_demux1(self, lls, dg)
XObj self, lls;
Msg  *dg;
{
    XTime now;
    char  buf[MESSAGE_LENGTH];
    Msg   rdg;

    xGetTime(&now);		/* TEMP */
    xTrace3(consul_timingp, TR_DETAILED,
	    "USER: entered user_demux1, got msg of len %d at %d.%d",
	    msgLen(dg), now.sec, now.usec);
    xIfTrace(consul_timingp, TR_DETAILED) {
	MSG_SHOW(dg);
    }
    msgPop(dg, stripMsg, (VOID *)buf, MESSAGE_LENGTH, (VOID *)NULL);
    xTrace3(consul_timingp, TR_DETAILED,
	    "USER: seqno = %d, partno = %d, will reply = %d",
	    *((int *)buf), *((int *)buf+1), REPLY_TO_MSG(buf, 1));
    /*
     * if my turn to reply
     *   create a new msg
     *   push it to DISPATCH
     */
    if (REPLY_TO_MSG(buf, 1)) {
	/* increment the sequence number and reply */
	*((int *)buf)  += 1;	/* increment count */
        *((int *)buf+1) = 1;	/* insert my participant number */
	msgConstructBuffer(&rdg, buf, MESSAGE_LENGTH);
	xTrace2(consul_timingp, TR_DETAILED, "USER: sending msg, TIME %d.%d",
                now.sec, now.usec);
	xIfTrace(consul_timingp, TR_DETAILED) {
	    MSG_SHOW(&rdg);
	}
	xPush(lls, &rdg);
	msgDestroy(&rdg);
    }
    return XK_SUCCESS;
}

/*
 * Called when the 2nd participant receives a msg from DISPATCH.
 */
static xkern_return_t user_demux2(self, lls, dg)
XObj self, lls;
Msg  *dg;
{
    XTime now;
    char  buf[MESSAGE_LENGTH];
    Msg   rdg;

    xGetTime(&now);		/* TEMP */
    xTrace3(consul_timingp, TR_DETAILED,
	    "USER: entered user_demux2, got msg of len %d at %d.%d",
            msgLen(dg), now.sec, now.usec);
    xIfTrace(consul_timingp, TR_DETAILED) {
	MSG_SHOW(dg);
    }
    msgPop(dg, stripMsg, (VOID *)buf, MESSAGE_LENGTH, (VOID *)NULL);
    xTrace3(consul_timingp, TR_DETAILED,
	    "USER: seqno = %d, partno = %d, will reply = %d",
	    *((int *)buf), *((int *)buf+1), REPLY_TO_MSG(buf, 2));
    /*
     * if my turn to reply
     *   create a new msg
     *   push it to DISPATCH
     */
    if (REPLY_TO_MSG(buf, 2)) {
	/* increment the sequence number and reply */
	*((int *)buf)  += 1;	/* increment count */
        *((int *)buf+1) = 2;	/* insert my participant number */
	msgConstructBuffer(&rdg, buf, MESSAGE_LENGTH);
	xTrace2(consul_timingp, TR_DETAILED, "USER: sending msg, TIME %d.%d",
                now.sec, now.usec);
	xIfTrace(consul_timingp, TR_DETAILED) {
	    MSG_SHOW(&rdg);
	}
	xPush(lls, &rdg);
	msgDestroy(&rdg);
    }
    return XK_SUCCESS;
}

static xkern_return_t user_open_done(self, lls, llp, hlpType)
XObj self, lls, llp, hlpType;
{
    xTrace1(consul_timingp, TR_FUNCTIONAL_TRACE,
	    "USER: entering user_open_done; psync open complete self = %x",
	    self);
    return XK_SUCCESS;
}

static void subtime(t1, t2, t3)
XTime *t1, *t2, *t3;
{
    t3->sec = t2->sec - t1->sec;
    t3->usec = t2->usec - t1->usec;
    if (t3->usec < 0) {
	t3->usec += 1000000;
	t3->sec -= 1;
    }
}

/*
 * Strip off the message and return in `hdr'.
 */
static long stripMsg(hdr, src, len, arg)
VOID *hdr, *arg;
char *src;
long len;
{
    int count, sender;

    bcopy(src, (char *)hdr, len);
    bcopy((char *)hdr, (char *)&count, sizeof(int));
    bcopy((char *)hdr + sizeof(int), (char *)&sender, sizeof(int));
    return len;
}
