%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

typedef struct _options_t {
    unsigned long spi;
    int host[4];
    int alg;
    int iv;
    int mode;
    int keylen;
    char *key;
}  options_t;

typedef struct _option_t {
    int op_id;
    options_t op_val;
}  option_t;

typedef struct _record_t {
    options_t local;
    options_t remote;
}  record_t;

typedef struct _integer_t {
    unsigned long val;
    char *str;
} integer_t;

int i, ac, mode=0;
char **av;
option_t   option;
options_t  options;
record_t   record;

%}

%union {
    int nothing;
    integer_t integer;
    int ipaddr[4];
}

%token <nothing> LOCAL REMOTE SPI HOST ALG IV MODE KEY KEY_LEN 
%token <nothing> MD5 DES DES3 DES3IN RSASIGN AH ESP ESPTUNNEL
%token <integer> INTEGER
%token <ipaddr>  IPADDR

%start records


%%



records :  record
	| records record
	;

record:	  localpart remotepart
	{  printRecord(); }
	| localpart remotepart ';'
	{  printRecord(); }
	| remotepart localpart
	{  printRecord(); }
	| remotepart localpart ';'
	{  printRecord(); }
	| error ';'
	;

localpart :	LOCAL options
	{ options_t *p;

	  p = &record.local;
	  p->spi  = options.spi;
	  p->alg  = options.alg;
	  p->iv   = options.iv;
	  p->mode = options.mode;
	  p->keylen = options.keylen;
	  p->key  = options.key;
	  for (i = 0; i < 4; i++)
	    p->host[i] = options.host[i];
	     
	}
	;

remotepart :	REMOTE options
	{ options_t *p;

	  p = &record.remote;
	  p->spi  = options.spi;
	  p->alg  = options.alg;
	  p->iv   = options.iv;
	  p->mode = options.mode;
	  p->key  = options.key;
	  p->keylen = options.keylen;
	  for (i = 0; i < 4; i++)
	    p->host[i] = options.host[i];
	}
	;

options :	option
	{ options.spi = -1;
	  for (i = 0; i < 4; i++)
	      options.host[i] = -1;
	  options.alg = -1;
	  options.iv = -1;
	  options.mode = -1;
	  options.keylen = -1;
	  options.key = 0;

	  switch (option.op_id) {
	      case SPI:   options.spi = option.op_val.spi;
			  break;
	      case HOST:  for (i = 0; i < 4; i++)
			      options.host[i] = option.op_val.host[i];
			  break;
	      case ALG:   options.alg = option.op_val.alg;
			  break;
	      case IV:    options.iv = option.op_val.iv;
			  break;
	      case MODE:  options.mode = option.op_val.mode;
			  break;
	      case KEY_LEN:options.keylen = option.op_val.keylen;
			  break;
	      case KEY:   options.key  = option.op_val.key;
			  break;
	  }
	}
	| options option
	{ switch (option.op_id) {
	      case SPI:   if (options.spi != -1) {
			      fprintf(stderr, "Cannot specify SPI twice\n");
			      YYERROR;
			  }
			  else
			      options.spi = option.op_val.spi;
			  break;
	      case HOST:  for (i = 0; i < 4; i++)
			      if (options.host[i] != -1) {
			           fprintf(stderr, "Cannot specify SPI twice\n");
				   YYERROR;
			      }
			      else
			          options.host[i] = option.op_val.host[i];
			  break;
	      case ALG:   if (options.alg != -1) {
			      fprintf(stderr, "Cannot specify ALG twice\n");
			      YYERROR;
			  }
			  else
			      options.alg = option.op_val.alg;
			  break;
	      case IV:    if (options.iv != -1) {
			       fprintf(stderr, "Cannot specify IV twice\n");
			       YYERROR;
			  }
			  else
			      options.iv = option.op_val.iv;
			  break;
	      case MODE:  if (options.mode != -1) {
			      fprintf(stderr, "Cannot specify MODE twice\n");
			      YYERROR;
			  }
			  else
			      options.mode = option.op_val.mode;
			  break;
	      case KEY_LEN:if (options.keylen != -1) {
			      fprintf(stderr, "Cannot specify KEY_LEN twice\n");
			      YYERROR;
			  }
			  else
			      options.keylen = option.op_val.keylen;
			  break;
	      case KEY:   if (options.key && strlen(options.key) != 0) {
			      fprintf(stderr, "Cannot specify KEY twice\n");
			      YYERROR;
			  }
			  else {
			      options.key = option.op_val.key;
			  }
			  break;
	  }
	}
	;

option :	SPI '=' INTEGER
	{ option.op_id = SPI;
	  option.op_val.spi = $3.val;
	}
	|	SPI '=' IPADDR
	{ option.op_id = SPI;
	  option.op_val.spi = 0;
	  for (i = 0; i < 4; i++)
	      if ($3[i] >= 0 && $3[i] <= 255)
		  option.op_val.spi = option.op_val.spi * 256 + $3[i];
	      else {
		  fprintf(stderr, "Invalid IP address\n");
		  YYERROR;
	      }
	}
	|	HOST '=' IPADDR
	{ option.op_id = HOST;
	  for (i = 0; i < 4; i++)
	      if ($3[i] >= 0 && $3[i] <= 255)
		  option.op_val.host[i] = $3[i];
	      else {
		  fprintf(stderr, "Invalid IP address\n");
		  YYERROR;
	      }
	}
	|	ALG '=' MD5
	{ option.op_id = ALG;
	  option.op_val.alg = MD5;
	}
	|	ALG '=' DES
	{ option.op_id = ALG;
	  option.op_val.alg = DES;
	}
	|	ALG '=' DES3
	{ option.op_id = ALG;
	  option.op_val.alg = DES3;
	}
	|	ALG '=' DES3IN
	{ option.op_id = ALG;
	  option.op_val.alg = DES3IN;
	}
	|	ALG '=' RSASIGN
	{ option.op_id = ALG;
	  option.op_val.alg = RSASIGN;
	}
	|	IV '=' INTEGER
	{ option.op_id = IV;
	  option.op_val.iv = $3.val;
	  if ($3.val != 0 && $3.val != 4 && $3.val != 8)
	      YYERROR;
	}
	|	MODE '=' AH
	{ option.op_id = MODE;
	  option.op_val.mode = AH;
	}
	|	MODE '=' ESP
	{ option.op_id = MODE;
	  option.op_val.mode = ESP;
	}
	|	MODE '=' ESPTUNNEL
	{ option.op_id = MODE;
	  option.op_val.mode = ESPTUNNEL;
	}
	|	KEY_LEN '=' INTEGER
	{ option.op_id = KEY_LEN;
	  option.op_val.keylen = 2 * $3.val;
	}
	|	KEY '=' INTEGER
	{ option.op_id = KEY;
	  option.op_val.key = $3.str;
	}
	;



%%		  				

		  				
int readNumber(char *str, integer_t *num)
{
  int i, a, b, c, d, tmp;
  char *cp, str1[65], str2[65];
  options_t *op;

  if (strcmp(str, "0") == 0) { /* 0 */
    num->val = 0;
    strcpy(num->str, "0");
  }
  else if (strncmp(str, "0x", 2) == 0) { /* hex number */
    num->val = 0;
    if (strlen(str) > 50) str[50] = '\0';
    for (i = 2; i < strlen(str); i++) {
      if (str[i] >= '0' && str[i] <= '9')
	num->val = num->val * 16 + str[i] - '0' ;
      else if (str[i] >= 'a' && str[i] <= 'f')
	num->val = num->val * 16 + str[i] - 'a' + 10;
      else if (str[i] >= 'A' && str[i] <= 'F')
	num->val = num->val * 16 + str[i] - 'A' + 10;
      else {
	fprintf(stderr, "Invalid number:%s--Ignored\n", str);
	return(-1);
      }
    }
    strcpy(num->str, str + 2);
#ifdef DEBUG
printf("%x\t", num->val);
printf("%s\n", num->str);
#endif
    return (0);
  }
  else if (*str == '0') { /* oct number */
    num->val = 0; str1[0] = str2[0] = '\0';
    if (strlen(str) > 65) str[65] = '\0';
    for (i = 1; i < strlen(str); i++) {
      if (str[i] >= '0' && str[i] <= '7')
	num->val = num->val * 8 + str[i] - '0';
      else {
	fprintf(stderr, "Invalid number:%s--Ignored\n", str);
	return(-1);
      }
    }
    for (i = strlen(str); i >= 1; i = i - 4) {
      a = (i   >= 1 ? str[i]   - '0' : 0);
      b = (i-1 >= 1 ? str[i-1] - '0' : 0);
      c = (i-2 >= 1 ? str[i-2] - '0' : 0);
      d = (i-3 >= 1 ? str[i-3] - '0' : 0);
      tmp = a + b*8 + c*64 + d*512;
      sprintf(str1, "%03x", tmp);
      strcat(str1, str2); strcpy(str2, str1);
    }
    for(i = 0; str2[i] == '0'; i++);
    if(str2[i] == '\0') i--;
    strcpy(num->str, str2);
    num->str[strlen(str2+i)] = '\0';
#ifdef DEBUG
printf("%o\t", num->val);
printf("%s\n", num->str);
#endif
    return(0);
  }
  else if (*str >= '1' && *str <= '9') { /* dec number */
    num->val = 0;
    for (i = 0; i < strlen(str); i++) {
      if (str[i] >= '0' && str[i] <= '9')
	num->val = num->val * 10 + str[i] - '0';
      else {
	fprintf(stderr, "Invalid number:%s--Ignored\n", str);
	return(-1);
      }
    }
    sprintf(num->str, "%x", num->val);
#ifdef DEBUG
printf("%u\t", num->val);
printf("%s\n", num->str);
#endif
    return(0);
  }
  else { /* string */
    return(readString(str, num));
  }
}


int readString(char *str, integer_t *num)
{ int i;

  num->val = 0;
  if (strlen(str) > 24) str[24] = '\0';
  for (i = 0; i < strlen(str); i++) {
    num->val = num->val * 256 + str[i];
    sprintf(num->str+2*i, "%02x", str[i]);
  }
#ifdef DEBUG
printf("%x\t", yylval.integer.val);
printf("%s\n", yylval.integer.str);
#endif
  return(0);
}
  

int readIPAddress(char *str, int ipaddr[4])
{ int i, j;
  
  for (j = 0; j < 4; j++)
    ipaddr[j] = 0;
  for (i = 0, j = 0; i < strlen(str) && j < 4; i++) {
    if (str[i] == '.') j++;
    else if (str[i] >= '0' && str[i] <= '9')
      ipaddr[j] = ipaddr[j] * 10 + str[i] - '0';
    else {
      fprintf(stderr, "Invalid IP address: %s--Ignored\n", str);
      return(-1);
    }
  }
  if (j >= 4) {
    fprintf(stderr, "Invalid IP address: %s--Ignored\n", str);
    return(-1);
  }
  for (j = 0; j < 4; j++) 
    if (ipaddr[j] < 0 || ipaddr[j] > 255) {
      fprintf(stderr, "Invalid IP address: %s--Ignored\n", str);
      return(-1);
    }
#ifdef DEBUG
for (i = 0; i < 4; i++) {
    printf("%d\t",ipaddr[i]);
}
printf("\n");
#endif
  return(0);
}


int readOptions(int argc, char **argv)
{  int i;
   char *cp;
   int ipaddr[4];
   integer_t num;
   options_t *op;
   
   mode = argc > 1 ? 1 : 0;
   
   for (i = 1; i < argc; i++) {
     if (strlen(argv[i]) < 2 || *(argv[i]) != '-') {
       fprintf(stderr, "Illegal command line option: %s--Ignored\n", argv[i]);
       continue;
     }

     if (strncmp(argv[i], "-l", 2) == 0) {
       op = &record.local;
       cp = argv[i] + 2;
     }
     else if (strncmp(argv[i], "-r", 2) == 0) {
       op = &record.remote;
       cp = argv[i] + 2;
     }
     else if (strcmp(argv[i], "-s") == 0) {
       mode = 0;
     }
     else if (strcmp(argv[i], "-p") == 0) {
       mode = 1;
     }
     else {
       fprintf(stderr, "Illegal command line option: %s--Ignored\n", argv[i]);
       continue;
     }

     switch (*cp++) {
       case 'a' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];

	          if (strcmp(cp, "MD5") == 0 || strcmp(cp, "md5") == 0) op->alg = MD5;
	          else if (strcmp(cp, "DES") == 0 || strcmp(cp, "des") == 0) op->alg = DES;
	          else if (strcmp(cp, "DES3") == 0 || strcmp(cp, "des3") == 0) op->alg = DES3;
	          else if (strcmp(cp, "DES3IN") == 0 || strcmp(cp, "des3in") == 0) op->alg = DES3IN;
	          else {
		    fprintf(stderr, "Illegal algorithm: %s--Ignored\n", cp);
		  }
	          break;

       case 'm' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];

	          if (strcmp(cp, "AH") == 0 || strcmp(cp, "ah") == 0) op->mode = AH;
	          else if (strcmp(cp, "ESP") == 0 || strcmp(cp, "esp") == 0) op->mode = ESP;
	          else if (strcmp(cp, "ESPTUNNEL") == 0 || strcmp(cp, "esptunnel") == 0) op->mode = ESPTUNNEL;
	          else {
		    fprintf(stderr, "Illegal mode: %s--Ignored\n", cp);
		  }
	          break;

       case 'i' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];

	          if (strcmp(cp, "0") == 0) op->iv = 0;
	          else if (strcmp(cp, "4") == 0) op->iv = 4;
	          else if (strcmp(cp, "8") == 0) op->iv = 8;
	          else {
		    fprintf(stderr, "Illegal iv: %s--Ignored\n", cp);
		  }
	          break;

       case 's' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];

	          if (readNumber(cp, &num) == 0) {
		    op->spi = num.val;
		  }
	          break;

       case 'h' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];
	    
	          if (readIPAddress(cp, &ipaddr[0]) > 0) {
		    for (i = 0; i < 4; i++)
		      op->host[i] = ipaddr[i];
		  }
	          break;

       case 'k' : if (*cp == '=') cp++;
	          else if (*cp == '\0') cp = argv[++i];

	          if (readNumber(cp, &num) == 0) {
		    strcpy(op->key, num.str);
		  }
	          break;
       default  : fprintf(stderr, "Illegal command line option: %s--Ignored\n", cp);
	          break;
     } 
  }
}


int printRecordPart(options_t *op)
{
   printf("%u;%d.%d.%d.%d;", op->spi, op->host[0], op->host[1], op->host[2], op->host[3]);
   switch (op->alg) {
       case MD5     :  printf("md5t;");   break;
       case DES     :  printf("des");     break;
       case DES3    :  printf("des3");    break;
       case DES3IN  :  printf("des3in");  break;
       case RSASIGN :  printf("rsasignt;"); break;
   }
   if (op->alg != MD5 && op->alg != RSASIGN)
       printf("%dt;",op->iv);
   printf("0;");
   switch (op->mode) {
       case AH          :  printf("0;");     break;
       case ESPTUNNEL   :  printf("1;");     break;
       case ESP         :  printf("2;");     break;
   }
   printf("%s", op->key);
}
  
int printRecord()
{  options_t *op;

   readOptions(ac, av);
   checkRecord(0);
   checkRecord(1);

   printRecordPart(&record.local);
   printf("=");
   printRecordPart(&record.remote);
   printf("\n");

   if (mode) {
     printRecordPart(&record.remote);
     printf("=");
     printRecordPart(&record.local);
     printf("\n");
   }
   
}


void sm_xk_des_setparity(unsigned char *key)
{  int i, k;
   for (i = 0; i < 8; i++) {
       k = 1 ^ key[i]; 
       k ^= k >> 4;
       k ^= k >> 2;
       k ^= k >> 1;
       if (k & 1) key[i] ^= 1;
   }
}


int sm_des_set_key_parity(char *key)
{  char pkey[8];

   strncpy(pkey, key, 8); /* bcopy(key, pkey, 8); */
   sm_xk_des_setparity(pkey);
   if (strncmp(key, pkey, 8)) {
       fprintf(stderr, "key %s has bad parity, fixed\n", key);
       strncpy(key, pkey, 8); /* bcopy(pkey, key, 8) */
   }
}


int checkRecord(int which)
{  int i, j;
   char tmp1[49], tmp2[49];	/* key buffers */
   int key[12];
   unsigned char a, b, key2[8];
   options_t *p;

   if(which == 0) p = &record.local;
   else p = &record.remote;

   if (p->spi == -1) {
       fprintf(stderr, "Must specify %s SPI\n", which ? "REMOTE" : "LOCAL");
       yyerror();
   }
   if (p->host[0] == -1) {
       fprintf(stderr, "Must specify %s HOST\n", which ? "REMOTE" : "LOCAL");
       yyerror();
   }
   if (p->alg == -1) {
       p->alg = MD5;
   }
   if ((p->alg == MD5) && (p->iv != -1)) {
       fprintf(stderr, "Warning: MD5 has no IV\n");
   }
   if (p->iv == -1) {
       p->iv = 8;
   }
   if (p->mode == -1) {
       if (which == 0) 
	   fprintf(stderr, "Warning: Should specify %s MODE\n", which ? "REMOTE" : "LOCAL");
       if (p->alg == MD5) p->mode = AH;
       else p->mode = ESP;
   }

   if (p->key == (char *)0 || p->key[0] == '\0') { /* no key */
       if (which) yyerror();
       else { /* generate 24 byte random key */
           fprintf(stderr, "Warning: Didnot specify LOCAL KEY, choose a random one\n"); 
           for (i = 0; i < 12; i++)
	     key[i] = rand();
	   for (i = 0; i < 24; i += 8)
	     sm_xk_des_setparity((unsigned char *)key+i);
	   p->key = (char *)malloc(49);
	   for (i = 0; i < 12; i++)
	     sprintf(p->key + 4 * i, "%04x", key[i]);
       }
   }
   if (p->alg == DES || p->alg == DES3 || p->alg == DES3IN) { /* DES algorithm: set parity of key */
     strcpy(tmp1, p->key);	/* preserve key in case p->alg == DES3 */

     for (i = 0; i < 8; i++)
       key2[i] = 0;
     for (i = strlen(p->key) - 1, j = 7; i >= 0 && j >= 0; ) { /* use the last 16 chars in p->key */
       if (i >= 0) {
	 a = p->key[i--];
	 if (a >= '0' && a <= '9') a = a -'0';
	 else if (a >= 'a' && a <= 'f') a = a - 'a' + 10;
	 else a = a - 'A' + 10;
       }
       else a = 0;
       if (i >= 0) {
	 b = p->key[i--];
	 if (b >= '0' && b <= '9') b = b -'0';
	 else if (b >= 'a' && b <= 'f') b = b - 'a' + 10;
	 else b = b - 'A' + 10;
       }
       else b = 0;
       key2[j--] = a + b * 16;
     }
     sm_des_set_key_parity(key2);
     for (i = 0; i < 8; i++)
       sprintf(p->key + 2 * i, "%02x", key2[i]);
   }

   if (p->alg == DES3 || p->alg == DES3IN) {	/* set parity of the first and second keys */
       strcpy(tmp2, p->key);	/* preserve the already set-parity-ed second key */
     
       /* processing the second key */
       for (i = 0; i < 8; i++)
	   key2[i] = 0;
       for (i = strlen(tmp1) - 16 - 1, j = 7; i >= 0 && j >= 0; ) { 
	   if (i >= 0) {
	       a = tmp1[i--];	/* use the key preserved in tmp1 */
	       if (a >= '0' && a <= '9') a = a -'0';
	       else if (a >= 'a' && a <= 'f') a = a - 'a' + 10;
	       else a = a - 'A' + 10;
	   }
	   else a = 0;
	   if (i >= 0) {
	       b = tmp1[i--];
	       if (b >= '0' && b <= '9') b = b -'0';
	       else if (b >= 'a' && b <= 'f') b = b - 'a' + 10;
	       else b = b - 'A' + 10;
	   }
	   else b = 0;
	   key2[j--] = a + b * 16;
       }
       sm_des_set_key_parity(key2);
       for (i = 0; i < 8; i++)
	   sprintf(p->key + 2 * i, "%02x", key2[i]);
       strcat(p->key, tmp2); 
       strcpy(tmp2, p->key); /* preserve the key in tmp2 */

       /* processing the first key */
       for (i = 0; i < 8; i++)
	   key2[i] = 0;
       for (i = strlen(tmp1) - 32 - 1, j = 7; i >= 0 && j >= 0; ) { 
	   if (i >= 0) {
	       a = tmp1[i--];	/* use the key preserved in tmp1 */
	       if (a >= '0' && a <= '9') a = a -'0';
	       else if (a >= 'a' && a <= 'f') a = a - 'a' + 10;
	       else a = a - 'A' + 10;
	   }
	   else a = 0;
	   if (i >= 0) {
	       b = tmp1[i--];
	       if (b >= '0' && b <= '9') b = b -'0';
	       else if (b >= 'a' && b <= 'f') b = b - 'a' + 10;
	       else b = b - 'A' + 10;
	   }
	   else b = 0;
	   key2[j--] = a + b * 16;
       }
       sm_des_set_key_parity(key2);
       for (i = 0; i < 8; i++)
	   sprintf(p->key + 2 * i, "%02x", key2[i]);
       strcat(p->key, tmp2); 
   }

   if (p->keylen == -1) {	/* default key length: 16 */
       p->keylen = 16;
   }
   if (strlen(p->key) > 0) p->keylen = strlen(p->key);
 }


main(int argc, char **argv)
{
  ac = argc;
  av = argv;

  yyparse();

  return(0);

}


	/*--------------------------------------------------------------
	  Since bison requires the user to implement yyerror we might
	  as well do it for all versions of yacc.
	  --------------------------------------------------------------*/
	  
int yyerror(char *s) {
  fprintf (stderr, "%s\n", s);
}
