/*
 * $RCSfile: icode.c,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: icode.c,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

/*
 * Basic functions for handling intermediate code
 */

#include "usc.h"
#include "icode.h"
#include "gen_utils.h"
#include "codegen_export.h"
#include "y.tab.h"

ICodeList *icode_list;
int temp_var_count;

ICodeList *new_icodelist (void)
{
  ICodeList *new = (ICodeList *)TempMalloc (sizeof (ICodeList));
  new->head = NULL;
  new->tail = NULL;
  new->num = 0;
  new->left_buckets.head = NULL;
  new->right_buckets.head = NULL;
  new->left_buckets.tail = NULL;
  new->right_buckets.tail = NULL;
  return new;
}

void insert_before (ICodeList *list, ICode *cur, ICode *new)
{
  if (cur->prev) {
    cur->prev->next = new;
    new->prev = cur->prev;
    new->next = cur;
    cur->prev = new;
  }
  else {
    list->head = new;
    new->prev = NULL;
    new->next = cur;
    cur->prev = new;
  }
}

void add_icode (ICodeList *list, ICode *new)
{
  if (list->head) {
    new->next = NULL;
    new->prev = list->tail;
    new->prev->next = new;
    list->tail = new;
  }
  else {
    new->next = NULL;
    new->prev = NULL;
    list->head = list->tail = new;
  }
  list->num++;
}

void push_icode (ICodeList *list, ICode *new)
{
  new->next = list->head;
  new->prev = NULL;
  list->head = new;
  if (!list->tail)
    list->tail = new;
  list->num++;
}

ICode *pop_icode (ICodeList *list)
{
  ICode *ans = list->head;
  if (list->head) {
    list->head = ans->next;
    if (list->head) {
      list->head->prev = NULL;
    }
    else {
      list->tail = NULL;
    }
    list->num--;
  }
  return ans;
}

void del_icode (ICodeList *list, ICode *old)
{
  if (old->prev) {
    old->prev->next = old->next;
  }
  else {
    list->head = old->next;
  }
  if (old->next) {
    old->next->prev = old->prev;
  }
  else {
    list->tail = old->prev;
  }
  list->num--;
}

ICode *new_icode (OpCode opcode)
{
  ICode *new = (ICode *)TempMalloc (sizeof (ICode));
  new->opcode = opcode;
  new->left.tempname = NULL;
  new->right.tempname = NULL;
  new->left.bucket = NULL;
  new->right.bucket = NULL;
  new->left.mask_head = NULL;
  new->right.mask_head = NULL;
  new->left.mask_tail = NULL;
  new->right.mask_tail = NULL;
  return new;
}

ICode *fix_icode_key (ICode *icode)
{
  Expr *offset = icode->left.offset;
  icode->left.key = 0;
  icode->right.key = 0;
  if (offset) {
    if (EXPR_IS_INT (offset))
      icode->left.key = EXPR_INT (offset);
    else if (EXPR_IS_PLUS (offset) && EXPR_IS_INT (EXPR_OP_LEFT (offset)))
      icode->left.key = EXPR_INT (EXPR_OP_LEFT (offset));
    else {
      printf ("%s\n", gen_offset (0, offset));
      crash ("internal error in fix_icode_key");
    }
  }

  offset = icode->right.offset;
  if (offset) {
    if (EXPR_IS_INT (offset))
      icode->right.key = EXPR_INT (offset);
    else if (EXPR_IS_PLUS (offset) && EXPR_IS_INT (EXPR_OP_LEFT (offset)))
      icode->right.key = EXPR_INT (EXPR_OP_LEFT (offset));
    else {
      printf ("%s\n", gen_offset (0, offset));
      crash ("internal error in fix_icode_key");
    }
  }
  return icode;
}

Expr *gen_temp (Expr *rvalue)
{
  char *tempname = (char *)TempMalloc (16 * sizeof (char));
  Expr *new = (Expr *)TempMalloc (sizeof (Expr));
  EXPR_TYPE (new) = IDENT_EXPR;
  sprintf (tempname, "temp%d", temp_var_count++);
  EXPR_IDENT (new) = tempname;
  gen_temp_icode (WORD_CODE, tempname, NULL,
                  WORD_CODE, NULL, rvalue);
  return new;
}

int icode_varcode_to_size (VarCode code)
{
  switch (code & ~(SIGN_FLAG | TEMP_FLAG | MASK_FLAG | ZERO_FLAG)) {
  case NULL_CODE:
	 /* return machine_info.long_size; */
	 return 4;
	 break;
	 
  case BYTE_CODE:
	 /* return machine_info.char_size; */
	 return 1;
	 
  case HALF_CODE:
	 /* return machine_info.short_size; */
	 return 2;
	 
  case WORD_CODE:
	 /* return machine_info.int_size; */
	 return 4;
	 
  case DOUB_CODE:
	 /* return machine_info.long_size; */
	 return 8;
	 
	 default:
	 crash ("internal error in icode_varcode_to_size: %d",
			  code & ~(SIGN_FLAG | TEMP_FLAG | MASK_FLAG | ZERO_FLAG));
	 break;
  }
  return 0;
}

ICode *new_temp_icode (VarCode leftvarcode, char *leftname, 
                       Expr *leftoffset, VarCode rightvarcode, 
                       char *rightname, Expr *rightoffset)
{
  ICode *new = new_icode (TEMP_OP);
  new->left.varcode = leftvarcode;
  new->left.name = allocate_temp_string (leftname);
  new->left.offset = leftoffset;
  new->right.varcode = rightvarcode;
  new->right.name = allocate_temp_string (rightname);
  new->right.offset = rightoffset;
  return new;
}

void gen_temp_icode (VarCode leftvarcode, char *leftname, 
                     Expr *leftoffset, VarCode rightvarcode, 
                     char *rightname, Expr *rightoffset)
{
  add_icode (icode_list, new_temp_icode (leftvarcode, leftname, leftoffset,
                                         rightvarcode, rightname, 
                                         rightoffset));
}

ICode *new_assign_icode (VarCode leftvarcode, char *leftname, 
			 Expr *leftoffset, Expr *real_leftoffset,
			 int leftalign, ICodeSide leftside, 
			 VarCode rightvarcode, char *rightname, 
			 Expr *rightoffset, Expr *real_rightoffset,
			 int rightalign, ICodeSide rightside)
{
  ICode *new = new_icode (ASSIGN_OP);
  new->left.key = 0;
  new->right.key = 0;
  new->left.varcode = leftvarcode;
  new->left.name = allocate_temp_string (leftname);
  new->left.offset = leftoffset;
  new->left.alignment = leftalign;
  new->left.real_offset = real_leftoffset;
  new->left.side = leftside;
  new->right.varcode = rightvarcode;
  new->right.name = allocate_temp_string (rightname);
  new->right.offset = rightoffset;
  new->right.real_offset = real_rightoffset;
  new->right.alignment = rightalign;
  new->right.side = rightside;
  return new;
}

void gen_assign_icode (VarCode leftvarcode, char *leftname, 
		       Expr *leftoffset, Expr *real_leftoffset,
		       int leftalign, ICodeSide leftside, 
		       VarCode rightvarcode, char *rightname, 
		       Expr *rightoffset, Expr *real_rightoffset,
		       int rightalign, ICodeSide rightside)
{
  ICode *new = new_assign_icode (leftvarcode, leftname,
				 leftoffset, real_leftoffset,
				 leftalign, leftside,
				 rightvarcode, rightname,
				 rightoffset, real_rightoffset,
				 rightalign, rightside);
  add_icode (icode_list, new);
}

void gen_assign_sign_icode (VarCode leftvarcode, char *leftname, 
                            Expr *leftoffset, int leftalign,
                            VarCode rightvarcode, char *tempname, 
                            char *rightname, Expr *rightoffset, 
                            int rightalign)
{
  ICode *new = new_icode (ASSIGN_OP);
  new->left.varcode = leftvarcode;
  new->left.name = allocate_temp_string (leftname);
  new->left.offset = leftoffset;
  new->left.alignment = leftalign;
  new->right.varcode = rightvarcode | TEMP_FLAG;
  new->right.name = allocate_temp_string (rightname);
  new->right.tempname = allocate_temp_string (tempname);
  new->right.offset = rightoffset;
  new->right.alignment = rightalign;
  add_icode (icode_list, new);
}

void gen_end_icode (void)
{
  add_icode (icode_list, new_icode (END_OP));
}

void gen_push_icode (void)
{
  add_icode (icode_list, new_icode (PUSH_OP));
}

void gen_pop_icode (void)
{
  add_icode (icode_list, new_icode (POP_OP));
}

void gen_for_icode (char *var, Expr *from, Expr *to)
{
  ICode *icode = new_icode (FOR_OP);
  icode->left.name = allocate_temp_string (var);
  icode->left.offset = from;
  icode->left.varcode = WORD_CODE;
  icode->right.offset = to;
  add_icode (icode_list, icode);
}

void gen_temp_return_icode (Expr *ret_temp)
{
  char* fmt;
  char name[128];
  ICode *icode = new_icode (RET_OP);
  if ( !EXPR_IS_SYMINFO(ret_temp) ) {
    crash("internal error in gen_temp_return_icode");
  }
  strcpy( name, symtype_to_string( EXPR_SYM(ret_temp) ) );
  if ( ( fmt = strchr(name,'%') ) != NULL )
    *fmt = '\0';
  icode->left.name = allocate_temp_string(name);
  icode->left.varcode = WORD_CODE;
  icode->right.name = SYM_LOCAL_NAME( EXPR_SYM(ret_temp) );
  icode->right.offset = ret_temp;
  add_icode (icode_list, icode);
}

void gen_return_icode( Expr *retval )
{
  ICode *icode = new_icode (RET_OP);
  icode->left.name = NULL;
  icode->left.varcode = WORD_CODE;
  icode->right.offset = retval;
  add_icode (icode_list, icode);
}

char *icodevarcode_to_string (VarCode varcode)
{
  static char s[128];
  s[0] = '\0';
  
  if (varcode & SIGN_FLAG)
    strcat (s, "sign ");
  if (varcode & TEMP_FLAG)
    strcat (s, "temp ");
  if (varcode & ZERO_FLAG)
	 strcat (s, "zero ");
  if (varcode & BYTE_CODE)
    strcat (s, "byte ");
  if (varcode & HALF_CODE)
    strcat (s, "half ");
  if (varcode & WORD_CODE)
    strcat (s, "word ");
  if (varcode & DOUB_CODE)
    strcat (s, "doub ");
  if (varcode & BIT_CODE)
	 strcat (s, "bit ");
  if ((varcode & ~(SIGN_FLAG | TEMP_FLAG | MASK_FLAG)) == NULL_CODE)
    strcat (s, "NULL");

  if (!s[0])
    crash ("internal error in icodevarcode_to_string: %d", varcode);

  return s;
}

void print_masks (Mask *list)
{
  while (list) {
    if (MASK_IS_ZERO (list)) {
      printf ("(0 & 0x%lx)", MASK_MASK (list));
    }
    else if (MASK_IS_SHIFT (list)) {
      printf ("(i%d << %d) & 0x%lx", MASK_BUCKET (list), MASK_SHIFT (list),
	      MASK_MASK (list));
    }
    else if (MASK_IS_SIGN (list)) {
      printf ("sign%d (i%d, %d) & 0x%lx", MASK_SIGN_BUCKET (list)->bucket_id,
	      MASK_BUCKET (list), MASK_SIGNBIT (list),
	      MASK_MASK (list));
    }
    else {
      printf ("?????");
    }
    list = MASK_NEXT (list);
    if (list)
      printf (" | ");
  }
  printf ("\n");
}

void print_single_icode (int level, ICode *icode)
{
  if (level <= Debug_Level) {
    if (icode->opcode == END_OP) {
      printf ("end\n");
    }
    else if (icode->opcode == FOR_OP) {
      printf ("for: (%s = %s; %s < %s; %s++)\n", icode->left.name, 
              gen_offset (0, icode->left.offset), icode->left.name,
              gen_offset (0, icode->right.offset), icode->left.name);
    }
    else if (icode->opcode == RET_OP) {
      printf ("return\n");
    }
    else if (icode->opcode == NAT_ASSGN_OP) {
      printf ("native assign: %s = ", icode->left.name );
      print_expr(level,0,icode->right.offset );
      printf ("\n");
    }
    else if (icode->opcode == PUSH_OP) {
      printf ("push\n");
    }
    else if (icode->opcode == POP_OP) {
      printf ("pop\n");
    }
    else if (icode->opcode == TEMP_OP) {
      printf ("temp: ");
      printf ("%s(%d)(", icodevarcode_to_string (icode->left.varcode),
              icode->left.alignment);
      if (icode->left.name) {
        printf ("%s", icode->left.name);
        printf ("[%s]", gen_offset (0, icode->left.offset));
      }
      printf ("), %s(%d)(", icodevarcode_to_string (icode->right.varcode),
              icode->right.alignment);
      if (icode->right.name && !(icode->right.varcode & TEMP_FLAG)) {
        printf ("%s", icode->right.name);
        printf ("[%s]", gen_offset (0, icode->right.offset));
      }
      else if (!icode->right.name && icode->right.offset) {
        printf ("%s", gen_offset (0, icode->right.offset));
      }
      printf (")\n");
    }
    else if (icode->opcode == ASSIGN_OP) {
      printf ("assign: keys = ");
      if (icode->left.bucket) {
	printf ("(%d %d), ", icode->left.key, icode->left.bucket->bucket_id);
      }
      else {
	printf ("(%d NL), ", icode->left.key);
      }
      if (icode->right.bucket) {
	printf ("(%d %d): ", icode->right.key, icode->right.bucket->bucket_id);
      }
      else {
	printf ("(%d NL): ", icode->right.key);
      }
      if (icode->left.varcode & MASK_FLAG) {
	printf ("o%d(%d) <- ", icode->left.key, icode->left.bucket->size);
	print_masks (icode->right.mask_head);
      }
      else {
	printf ("%s(%d)(", icodevarcode_to_string (icode->left.varcode),
		icode->left.alignment);
	if (icode->left.name) {
	  printf ("%s", icode->left.name);
	  printf ("[%s %s]", gen_offset (0, icode->left.offset),
		  gen_offset (0, icode->left.real_offset));
	}
	if (icode->right.varcode != NULL_CODE) {
	  char *name_string = NULL;
	  if (icode->right.side == RIGHT_SIDE)
	    name_string = icode->right.name;
	  else if (icode->right.side == LEFT_SIDE)
	    name_string = icode->left.name;
	  else
	    crash ("This icode's right side's been shot off!\n");

	  printf ("), %s(%d)(", icodevarcode_to_string
		  (icode->right.varcode),
		  icode->right.alignment);
	  if (icode->right.name && !(icode->right.varcode & TEMP_FLAG)) {
	    printf ("%s", name_string);
	    printf ("[%s %s]", gen_offset (0, icode->right.offset),
		    gen_offset (0, icode->right.real_offset));
	  }
	  else if (icode->right.varcode & TEMP_FLAG) {
	    printf ("%s = %s[%s %s]", icode->right.tempname, 
		    name_string,
		    gen_offset (0, icode->right.offset),
		    gen_offset (0, icode->right.real_offset));
	  }
	  printf (")\n");
	}
	else {
	  printf ("), NULL_CODE\n");
	}
      }
    }
    else {
      crash ("internal error in print_single_icode");
    }
  }
}

void print_code (int level, ICodeList *icode_list)
{
  ICode *icode = icode_list->head;
  if (level <= Debug_Level) {
	 printf ("------------------------------\n");
    printf ("ICode: Num: %d\n", icode_list->num);
    while (icode) {
      print_single_icode (level, icode);
      icode = icode->next;
    }
    printf ("\n");
  }
}



void print_buckets (int level)
{
  RegBucket *p;

  if (level <= Debug_Level) {
    printf ("Left buckets:\n");
    for (p = icode_list->left_buckets.head; p; p = BUCKET_NEXT (p)) {
      if (BUCKET_IS_PLAIN (p))
	printf ("plain bucket\n");
      else if (BUCKET_IS_SIGN (p))
	printf ("sign bucket: bucket %d bit %d\n", BUCKET_SIGN_BUCKET (p),
		BUCKET_SIGN_BIT (p));
      else
	printf ("I don't know what kind of bucket this is\n");
      printf ("id: %d\n", BUCKET_ID (p));
      if (BUCKET_SIDE (p) == NULL_SIDE)
	printf ("side: NULL\n");
      else if (BUCKET_SIDE (p) == LEFT_SIDE)
	printf ("side: left\n");
      else if (BUCKET_SIDE (p) == RIGHT_SIDE)
	printf ("side: right\n");
      printf ("name: %s\n", BUCKET_NAME (p));
      printf ("size: %d\n", BUCKET_SIZE (p));
      printf ("start: ");
      print_expr (level, 0, BUCKET_START (p));
      printf ("\n\n");
    }
    printf ("Right buckets:\n");
    for (p = icode_list->right_buckets.head; p; p = BUCKET_NEXT (p)) {
      if (BUCKET_IS_PLAIN (p))
	printf ("plain bucket\n");
      else if (BUCKET_IS_SIGN (p))
	printf ("sign bucket: bucket %d bit %d\n", BUCKET_SIGN_BUCKET (p),
		BUCKET_SIGN_BIT (p));
      else
	printf ("I don't know what kind of bucket this is\n");
      printf ("id: %d\n", BUCKET_ID (p));
      if (BUCKET_SIDE (p) == NULL_SIDE)
	printf ("side: NULL\n");
      else if (BUCKET_SIDE (p) == LEFT_SIDE)
	printf ("side: left\n");
      else if (BUCKET_SIDE (p) == RIGHT_SIDE)
	printf ("side: right\n");
      printf ("name: %s\n", BUCKET_NAME (p));
      printf ("size: %d\n", BUCKET_SIZE (p));
      printf ("start: ");
      print_expr (level, 0, BUCKET_START (p));
      printf ("\n\n");
    }
  }
}
