/*
 * $RCSfile: icode.h,v $
 *
 * Universal Stub Compiler
 *
 * Copyright (c) 1994    Arizona Board of Regents
 *
 * $Log: icode.h,v $
 * Revision 1.1  1994/10/13 17:43:30  dorgival
 * Initial revision
 *
 */

#ifndef _INICODEH
#define _INICODEH

typedef enum {
  NULL_SIDE,
  LEFT_SIDE,
  RIGHT_SIDE,
  MAX_SIDE       /* used as a sentinel in the sorting. no real var should
		    ever have this value */
} ICodeSide;

typedef enum { 
  NULL_CODE = 0,
  BYTE_CODE = 1,
  HALF_CODE = 2,
  WORD_CODE = 4,
  DOUB_CODE = 8,
  BIT_CODE  = 16,

  MASK_FLAG = 32,
  TEMP_FLAG = 64,
  SIGN_FLAG = 128,
  ZERO_FLAG = 256
} VarCode;

typedef enum {
  NULL_OP, 
  ASSIGN_OP,
  TEMP_OP, 
  PUSH_OP,
  POP_OP,
  FOR_OP,
  END_OP,
  RET_OP,
  NAT_ASSGN_OP
} OpCode;

typedef enum {
  SHIFT_MASK,
  SIGN_MASK,
  ZERO_MASK
} MaskType;

typedef enum {
  PLAIN_BUCKET,
  SIGN_BUCKET
} BucketType;

typedef struct reg_buck {
  ICodeSide side;   /* left or right */
  char *name;       /* name of the var we're getting this from */
  BucketType type;
  int bucket_id;    /* list is sorted by this */
  int size;         /* in bits */
  Expr *start;      /* in bits during optimization and in bytes during gen */
  
  int sign_bucket;  /* if it's a sign bucket, which bucket */
  int sign_bit;     /* if it's a sign bucket, which bit?   */

  struct reg_buck *next;
} RegBucket;

#define BUCKET_TYPE(a)         ((a)->type)
#define BUCKET_ID(a)           ((a)->bucket_id)
#define BUCKET_SIZE(a)         ((a)->size)
#define BUCKET_SIDE(a)         ((a)->side)
#define BUCKET_START(a)        ((a)->start)
#define BUCKET_NAME(a)         ((a)->name)
#define BUCKET_NEXT(a)         ((a)->next)
#define BUCKET_SIGN_BUCKET(a)  ((a)->sign_bucket)
#define BUCKET_SIGN_BIT(a)     ((a)->sign_bit)
#define BUCKET_IS_PLAIN(a)     ((a)->type == PLAIN_BUCKET)
#define BUCKET_IS_SIGN(a)      ((a)->type == SIGN_BUCKET)

typedef struct {
  RegBucket *head, *tail;
} RegBucketList;

typedef struct mask {   
  MaskType type;
  int bucket;
  int shift;              /* a left shift. if < 0 then a right shift       */
  int sign_bit;           /* the source bit                                */
  unsigned long mask;     /* shift then mask. for zero masks, ignore shift */
  RegBucket *sign_buck;   /* the bucket that the sign bits are in          */
  struct mask *next, *prev;
} Mask;

#define MASK_TYPE(a)          ((a)->type)
#define MASK_BUCKET(a)        ((a)->bucket)
#define MASK_SHIFT(a)         ((a)->shift)
#define MASK_SIGNBIT(a)       ((a)->sign_bit)
#define MASK_MASK(a)          ((a)->mask)
#define MASK_NEXT(a)          ((a)->next)
#define MASK_PREV(a)          ((a)->prev)
#define SET_MASK_SHIFT(a,b)   (MASK_TYPE(a) = SHIFT_MASK, MASK_SHIFT(a)=(b))
#define SET_MASK_SIGNBIT(a,b) (MASK_TYPE(a) = SIGN_MASK, MASK_SIGNBIT(a)=(b))
#define SET_MASK_ZERO(a)      (MASK_TYPE(a) = ZERO_MASK, MASK_BUCKET(a) = -1)
#define MASK_IS_SHIFT(a)      (MASK_TYPE(a) == SHIFT_MASK)
#define MASK_IS_SIGN(a)       (MASK_TYPE(a) == SIGN_MASK)
#define MASK_IS_ZERO(a)       (MASK_TYPE(a) == ZERO_MASK)
#define MASK_SIGN_BUCKET(a)   ((a)->sign_buck)

typedef struct {
  VarCode varcode;
  ICodeSide side;      /* LEFT_SIDE for lvalue. NOT whether this is the
			  ICode.left field. */
  char *name;
  int alignment;
  int key;             /* in bits at the start and then the bucket key */
  RegBucket *bucket;
  Expr *offset;                 /* with temps, suitable for generation */
  Expr *real_offset;            /* no temps in here, sir               */
  Mask *mask_head, *mask_tail;  /* will only be used by the right side */
  char *tempname;
} IVar;

typedef struct icode {
  OpCode opcode;
  IVar left, right;
  struct icode *next, *prev;
} ICode;

typedef struct {
  int num;
  ICode *head;
  ICode *tail;
  RegBucketList left_buckets, right_buckets;
} ICodeList;

#define ICODE_IS_END(a)    (!(a) || (a)->opcode == END_OP)
#define ICODE_IS_ASSIGN(a) ((a)->opcode == ASSIGN_OP)
#define ICODE_IS_ZERO(a)   (ICODE_IS_ASSIGN(a) && \
				 (a)->right.varcode & ZERO_FLAG)

extern char *icodevarcode_to_string (VarCode varcode);
extern void print_single_icode      (int level, ICode *icode);
extern ICode *new_sign_icode        (char *basename, Expr *sign_offset);
extern ICode *new_icode             (OpCode opcode);
extern ICodeList *new_icodelist     (void);
extern ICode *pop_icode             (ICodeList *list);
extern ICode *fix_icode_key         (ICode *icode);
extern ICode *new_temp_icode        (VarCode leftvarcode, char *leftname, 
				     Expr *leftoffset, VarCode rightvarcode, 
				     char *rightname, Expr *rightoffset);

extern void insert_before           (ICodeList *list, ICode *cur, ICode *new);
extern void add_icode               (ICodeList *list, ICode *new);
extern void push_icode              (ICodeList *list, ICode *new);
extern void del_icode               (ICodeList *list, ICode *old);
extern void gen_assign_icode        (VarCode leftvarcode, char *leftname, 
				     Expr *leftoffset, Expr *real_leftoffset,
				     int leftalign, ICodeSide leftside,
				     VarCode rightvarcode, char *rightname, 
				     Expr *rightoffset, Expr *real_rightoffset,
				     int rightalig, ICodeSide rightside);

extern ICode *new_assign_icode      (VarCode leftvarcode, char *leftname, 
				     Expr *leftoffset, Expr *real_leftoffset,
				     int leftalign, ICodeSide leftside,
				     VarCode rightvarcode, char *rightname, 
				     Expr *rightoffset, Expr *real_rightoffset,
				     int rightalig, ICodeSide rightside);

extern void gen_assign_sign_icode (VarCode leftvarcode, char *leftname, 
                                   Expr *leftoffset, int leftalign,
                                   VarCode rightvarcode, char *tempname,
                                   char *rightname, 
                                   Expr *rightoffset, int rightalign);

extern void gen_temp_icode        (VarCode leftvarcode, char *leftname, 
				   Expr *leftoffset, VarCode rightvarcode, 
				   char *rightname, Expr *rightoffset);

extern void gen_push_icode        (void);
extern void gen_pop_icode         (void);
extern Expr *gen_temp             (Expr *rvalue);
extern void gen_end_icode         (void);
extern int icode_varcode_to_size  (VarCode code);
extern void gen_for_icode         (char *var, Expr *from, Expr *to);
extern void gen_return_icode      (Expr *statement);
extern void gen_temp_return_icode (Expr *statement);
extern Expr *make_ident_expr      (char *string);
extern void init_icode            (void);
extern int was_bitfield_seen      (void);
extern void fix_bitfields         (Expr *left_expr);

extern void print_code (int level, ICodeList *icode_list);
extern void print_buckets (int level);


#endif
